package main;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import msgs.Message;
import msgs.MessageFactory;

public class Connection 
{
	//--------------------------------------------------------------------
	private Socket socket;
	private boolean connectionWorks = true;
	private ListenThread listener;
	private OutputStream output;
	
	private List<Message> receivedMsgs = Collections.synchronizedList(new LinkedList<Message>());	
	
	//--------------------------------------------------------------------
	public Connection (Socket socket) 
	{
		this.socket = socket;
		try 
		{
			output = socket.getOutputStream();
		} 
		catch (IOException e) 
		{
			System.out.println ("Connection: Can't get Output Stream: " + e);			
		}
		listener = new ListenThread (socket, receivedMsgs);
		listener.start ();
	}
	
	//--------------------------------------------------------------------
	public boolean isConnected () 
	{
		return (socket.isConnected () && connectionWorks);
		// TODO: send heartbeats?
	}	
	
	//--------------------------------------------------------------------
	public boolean msgAvailable() 
	{
		return receivedMsgs.size () > 0;
	}
	
	//--------------------------------------------------------------------
	public Message getNextMsg () 
	{
		if (msgAvailable ()) 
		{
			Message msg = receivedMsgs.get (0);
			receivedMsgs.remove (0);
			return msg;
		}
		return null; 
	}

	//--------------------------------------------------------------------
	public void sendMsg (Message msg) 
	{
		try 
		{
			synchronized (this) 
			{
				if (msg.writeToStream (output) == false)
					System.out.println ("Warning: Could not write message to stream: " + msg);
			}
			output.flush ();
		} 
		catch (IOException e) 
		{
			e.printStackTrace ();
			connectionWorks = false;
		} 
	}
	
	//--------------------------------------------------------------------
	public void close () 
	{
		if (listener.isListening ()) 
		{
			listener.close ();
			try 
			{
				listener.join ();
			} 
			catch (InterruptedException e) 
			{
				System.out.println ("Connection: Can't join listener Thread: " + e);
			}
		}
		try 
		{
			output.close ();
		} 
		catch (IOException e) 
		{
			System.out.println ("Connection: Can't close Output Stream: " + e);
		}	
		if (!socket.isClosed ()) 
		{
			try 
			{
				socket.close ();
			} 
			catch (IOException e) 
			{
				System.out.println ("Connection: Problems while closing socket: " + e);
			}
		}
	}

	//--------------------------------------------------------------------
	public void closeConnectionInThread () 
	{
		Thread t = new Thread () 
		{
			public void run () 
			{
				close ();
			}
		};
		t.run ();
	}
	
	
	/* ******************************************************************************* */
	/* ****************************** INNER CLASSES ********************************** */
	/* ******************************************************************************* */
	
	//--------------------------------------------------------------------
	class ListenThread extends Thread 
	{
	
		//--------------------------------------------------------------------
		private List<Message> msgList;
		private boolean listening = true;
		private InputStream input;

		//--------------------------------------------------------------------
		public ListenThread (Socket socket, List<Message> msgList) 
		{
			this.msgList = msgList;
			try 
			{
				input = socket.getInputStream ();
			} 
			catch (IOException e) 
			{
				System.out.println ("Connection: Can't get Input Stream: " + e);
			}
		}
		
		//--------------------------------------------------------------------
		public boolean isListening () {
			return listening;
		}
		
		//--------------------------------------------------------------------
		public void close () {
			listening = false;
		}
		
		//--------------------------------------------------------------------
		public void run () 
		{
			while (listening) 
			{
				try 
				{
					if (input.available () > 0) 
					{
						ArrayList<Message> resultMsgs = MessageFactory.createMsgsFromStream (input);
						if (resultMsgs != null && resultMsgs.size () > 0)
							msgList.addAll (resultMsgs);
					} 
					else 
					{
						try { sleep(50); } 
						catch (InterruptedException e)
						{
							System.out.println ("ListenThread: Problems reading Msg: " + e);
						}
					}
				} 
				catch (IOException e) 
				{
					e.printStackTrace();
				}
			}
			try 
			{
				input.close();
			} 
			catch (IOException e) 
			{
				System.out.println ("ClientConnection: Can't close Input Stream: " + e);
			}
		}
		
	} // end of ListenThread Class
	
}
