package model;

import java.awt.Point;
import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import msgs.ByteConvert;

//--------------------------------------------------------------------
public class Map 
{

	//--------------------------------------------------------------------
	public static final byte TT_NORMAL  = 0;
	public static final byte TT_WATER   = 1;
	public static final byte TT_COAST   = 2; 
	public static final byte TT_BLOCKED = 3;

	public static final byte RT_UNKNOWN = 0;
	public static final byte RT_NONE = 1;
	public static final byte RT_RAW  = 2;
	public static final byte RT_OIL  = 3; 
	public static final byte RT_GOLD = 4;

	//--------------------------------------------------------------------
	public Map (int width, int height)
	{
		this.width = width;
		this.height = height;
		terrainInfo = new byte[width * height];
		
		resourceTypeInfo = new byte[width * height];
		Arrays.fill (resourceTypeInfo, RT_UNKNOWN);
		resourceAmmountInfo = new byte[width * height];
		Arrays.fill (resourceAmmountInfo, (byte)0);
	}
	
	//--------------------------------------------------------------------
	public int getWidth () { return width; }
	public int getHeight () { return height; }
	public int getMapOffset (int x, int y) { return x + y * width; }
	public int getXFromOffset (int offset) { return offset % width; }
	public int getYFromOffset (int offset) { return offset / width; }
	public boolean areValidCoords (int x, int y) { return (x >= 0 && x < width && y >= 0 && y < height); }
	public boolean isValidOffset (int offset) { return (offset >= 0 && offset < width * height); }
	//--------------------------------------------------------------------
	private int width;
	private int height;

	
	//--------------------------------------------------------------------
	private byte[] terrainInfo = null;
	//--------------------------------------------------------------------
	public byte getTerrain (int x, int y)
	{
		if (areValidCoords (x, y))
			return terrainInfo[getMapOffset (x, y)];
		return TT_NORMAL;
	}	
	//--------------------------------------------------------------------
	private void setTerrain (int x, int y, byte terrain)
	{
		assert (areValidCoords (x, y));
		terrainInfo[getMapOffset (x, y)] = terrain;
	}
	
	//--------------------------------------------------------------------
	private byte[] resourceTypeInfo = null;	
	private byte[] resourceAmmountInfo = null;	
	//--------------------------------------------------------------------
	public boolean isResourceExplored (int x, int y) { return (getResourceType (x, y) != RT_UNKNOWN); }
	//--------------------------------------------------------------------
	public byte getResourceType (int x, int y)
	{
		if (areValidCoords (x, y))
			return resourceTypeInfo[getMapOffset (x, y)];
		return RT_UNKNOWN;
	}
	//--------------------------------------------------------------------
	public byte getResourceAmmount (int x, int y)
	{
		if (areValidCoords (x, y))
			return resourceAmmountInfo[getMapOffset (x, y)];
		return 0;
	}
	//--------------------------------------------------------------------
	public void setResource (int x, int y, byte type, byte ammount)
	{
		setResourceType (x, y, type);
		setResourceAmmount (x, y, ammount);
	}
	//--------------------------------------------------------------------
	public void setResourceType (int x, int y, byte type)
	{
		if (areValidCoords (x, y) && RT_UNKNOWN <= type && type <= RT_GOLD)
			resourceTypeInfo[getMapOffset (x, y)] = type;
	}
	//--------------------------------------------------------------------
	public void setResourceAmmount (int x, int y, byte ammount)
	{
		if (areValidCoords (x, y))
			resourceAmmountInfo[getMapOffset (x, y)] = ammount;		
	}
	
	
	//--------------------------------------------------------------------
	// C O N T I N E N T     A N A L Y S I S
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	private short[] continentMap = null;
	private ArrayList<ContinentInfo> continents = null;
	//--------------------------------------------------------------------
	class ContinentInfo
	{
		public ContinentInfo (int idx, boolean ocean)
		{
			this.idx = idx;
			this.ocean = ocean;
			size = 0;
		}
		public int idx;
		public boolean ocean;
		public int size;
	}
	//--------------------------------------------------------------------
	public void analyzeMap ()
	{
		continents = new ArrayList<ContinentInfo> ();
		continentMap = new short[width * height];
		for (int i = 0; i < width * height; i++)
			continentMap[i] = -2;
				
		short continentIdx = 0;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (continentMap[getMapOffset (x, y)] == -2)
				{
					if (getTerrain (x, y) == TT_BLOCKED)
						continentMap[getMapOffset (x, y)] = -1;
					else
					{
						analyzeNewContinent (continentIdx, x, y);
						continentIdx++;
					}
				}
			}
		}		
	}
	
	//--------------------------------------------------------------------
	private void analyzeNewContinent (short continentIdx, int x, int y)
	{		
		boolean isOcean = (getTerrain (x, y) == TT_WATER);
		ContinentInfo curContinentInfo = new ContinentInfo (continentIdx, isOcean);
		continents.add (curContinentInfo);

		ArrayList<SimplePoint> addedPositions = new ArrayList<SimplePoint> ();
		addedPositions.add (new SimplePoint (x, y));
		for (int i = 0; i < addedPositions.size (); i++)
		{
			SimplePoint curPoint = addedPositions.get (i);
			byte terrain = getTerrain (curPoint.x, curPoint.y);
			if ((terrain == TT_WATER && isOcean) || (!isOcean && (terrain == TT_NORMAL)))
			{
				continentMap [getMapOffset (curPoint.x, curPoint.y)] = continentIdx;
				curContinentInfo.size++;
			}
			else if (terrain == TT_BLOCKED || terrain == TT_COAST)
			{
				continentMap [getMapOffset (curPoint.x, curPoint.y)] = -1;
				continue;
			}
			else
				continue;

			if (curPoint.x > 0 && continentMap [getMapOffset (curPoint.x - 1, curPoint.y)] == -2) // left of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x - 1, curPoint.y));
				continentMap [getMapOffset (curPoint.x - 1, curPoint.y)] = -3;
			}
			if (curPoint.x + 1 < width && continentMap [getMapOffset (curPoint.x + 1, curPoint.y)] == -2) // right of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x + 1, curPoint.y));
				continentMap [getMapOffset (curPoint.x + 1, curPoint.y)]  = -3;
			}
			if (curPoint.y > 0 && continentMap [getMapOffset (curPoint.x, curPoint.y - 1)] == -2) // top of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x, curPoint.y - 1));
				continentMap [getMapOffset (curPoint.x, curPoint.y - 1)]  = -3;
			}
			if (curPoint.y + 1 < height && continentMap [getMapOffset (curPoint.x, curPoint.y + 1)] == -2) // under current point
			{
				addedPositions.add (new SimplePoint (curPoint.x, curPoint.y + 1));
				continentMap [getMapOffset (curPoint.x, curPoint.y + 1)]  = -3;
			}
			if (curPoint.x > 0 && curPoint.y > 0 && continentMap [getMapOffset (curPoint.x - 1, curPoint.y - 1)] == -2) // upper left of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x - 1, curPoint.y - 1));
				continentMap [getMapOffset (curPoint.x - 1, curPoint.y - 1)] = -3;
			}
			if (curPoint.x + 1 < width && curPoint.y > 0 && continentMap [getMapOffset (curPoint.x + 1, curPoint.y - 1)] == -2) // upper right of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x + 1, curPoint.y - 1));
				continentMap [getMapOffset (curPoint.x + 1, curPoint.y - 1)] = -3;
			}
			if (curPoint.x > 0 && curPoint.y + 1 < height && continentMap [getMapOffset (curPoint.x - 1, curPoint.y + 1)] == -2) // lower left of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x - 1, curPoint.y+1));
				continentMap [getMapOffset (curPoint.x - 1, curPoint.y + 1)] = -3;
			}
			if (curPoint.x + 1 < width && curPoint.y + 1 < height && continentMap [getMapOffset (curPoint.x + 1, curPoint.y + 1)] == -2) // lower right of current point
			{
				addedPositions.add (new SimplePoint (curPoint.x + 1, curPoint.y + 1));
				continentMap [getMapOffset (curPoint.x + 1, curPoint.y + 1)] = -3;
			}
		}
		// reset points, that don't belong to this continent
		for (int i = 0; i < addedPositions.size (); i++)
		{
			SimplePoint curPoint = addedPositions.get (i);
			if (continentMap [getMapOffset (curPoint.x, curPoint.y)] == -3)
				continentMap [getMapOffset (curPoint.x, curPoint.y)] = -2;
		}
	}
	//--------------------------------------------------------------------
	class SimplePoint
	{
		SimplePoint (int x, int y)
		{
			this.x = x;
			this.y = y;
		}
		public int x;
		public int y;
	}
	

	//--------------------------------------------------------------------
	// S T A R T I N G   L O C A T I O N   A N A L Y I S
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	public Point findStartingPlace ()
	{
		if (continentMap == null)
			analyzeMap ();
		
		Random randomizer = new Random ();
		int x = 0;		
		int y = 0;
		int bestCandidateX = 0;
		int bestCandidateY = 0;
		int bestRating = -1;
		int count = 0;
		while (bestRating < 0 || count < 50)
		{
			x = randomizer.nextInt (width);
			y = randomizer.nextInt (height);
			
			int rating = calcRating (x, y);
			if (rating > 0)
				count++;
			if (rating > bestRating)
			{
				bestRating = rating;
				bestCandidateX = x;
				bestCandidateY = y;
			}
		}
		return new Point (bestCandidateX, bestCandidateY);
	}

	//--------------------------------------------------------------------
	private int calcRating (int x, int y)
	{
		if (getTerrain(x, y) != TT_NORMAL)
			return -1;
		ContinentInfo continent = continents.get (continentMap[getMapOffset (x, y)]);
		if (continent.ocean)
			return -1;
		
		int rating = 50;
		if (continent.size < 30) // very small continent
			rating -= 10;
		else if (continent.size > 60) // big enough continent
			rating += 10;
		if (x < 5 || x >= width - 5) // don't get too near to the edge of the map
			rating -= 5;
		if (y < 5 || y >= height - 5) // don't get too near to the edge of the map
			rating -= 5;
		
		int[] rayLength = {0,0,0,0,0,0,0,0};
		int rayX = x; int rayY = y;
		while (areValidCoords (rayX, rayY) // to the right
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[0]++;
			rayX++;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the lower right
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[1]++;
			rayX++;
			rayY++;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the bottom
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[2]++;
			rayY++;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the lower left
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[3]++;
			rayX--;
			rayY++;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the left
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[4]++;
			rayX--;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the upper left
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[5]++;
			rayX--;
			rayY--;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the top
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[6]++;
			rayY--;
		}
		rayX = x; rayY = y;	
		while (areValidCoords (rayX, rayY) // to the upper right
				&& continentMap[getMapOffset (rayX, rayY)] == continent.idx)
		{
			rayLength[7]++;
			rayX++;
			rayY--;
		}
		Arrays.sort (rayLength);
		int smallest3RaysSum = rayLength[0] + rayLength[1] + rayLength[2];
		if (smallest3RaysSum < 10 || 20 < smallest3RaysSum) // quite, but not too close to the continent border 
			rating -= 4;
		if (rayLength[0] < 3) // very close to the continent border
			rating -= 3;

		int raysSum = 0;
		for (int i = 0; i < 8; i++)
			raysSum += rayLength[i];
		if (raysSum < 40) // too few expansion space
			rating -= 5;
		else if (80 < raysSum) // too much space to defend?
			rating -= 3;
		
		System.out.println ("Start location rating: " + rating + "  X: " + x + "  Y: " + y);
		return rating;
	}
	
	
	//--------------------------------------------------------------------
	// D E B U G 
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	public void printMap ()
	{
		System.out.println ("Terrain:");
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				System.out.print (getTerrain (x, y));
			}
			System.out.println ("");
		}
		if (continentMap != null)
		{
			System.out.println ("Continents:");
			for (int y = 0; y < height; y++)
			{
				for (int x = 0; x < width; x++)
				{
					if (continentMap[getMapOffset (x, y)] == -1)
						System.out.print(" x");
					else if (continentMap[getMapOffset (x, y)] < 10)
						System.out.print (" " + continentMap[getMapOffset (x, y)]);
					else
						System.out.print (continentMap[getMapOffset (x, y)]);
				}
				System.out.println ("");
			}
		}
	}

	
	//--------------------------------------------------------------------	
	// L O A D I N G
	//--------------------------------------------------------------------
	
	//--------------------------------------------------------------------
	public static Map loadMap (File file)
	{
		try 
		{
			FileInputStream fileStream = new FileInputStream (file);
			byte[] rawData = new byte[fileStream.available ()];
			int bytesRead = fileStream.read (rawData);
			if (bytesRead != rawData.length)
				System.out.println ("Error Map.loadMap: couldn't read the map file completely.");			
			fileStream.close ();

			String type = new String (rawData, 0, 3);
			if (!(type.equals ("WRL") || type.equals ("WRX") || type.equals ("DMO")))
			{
				System.out.println ("Error: Wrong map type: " + type);
				return null;
			}
			
			int curPos = 5;
			int width = ByteConvert.toInt16 (rawData, curPos);
			curPos += 2;
			int height = ByteConvert.toInt16 (rawData, curPos);
			curPos += 2;
			curPos += width * height;
			int dataPos = curPos;
			curPos += 2 * width * height;
			int numberOfTerrains = ByteConvert.toInt16 (rawData, curPos);
			curPos += 2;
			int graphicsPos = curPos;
			int palettePos = graphicsPos + numberOfTerrains * 64 * 64;
			int infoPos = palettePos + 256 * 3;
			
			// Load necessary Terrain Graphics
			byte[] terrainTypes = new byte[numberOfTerrains];
			for (int i = 0; i < numberOfTerrains; i++)
			{
				switch (rawData[infoPos + i])
				{
				case 0:
					//normal terrain without special property
					terrainTypes[i] = TT_NORMAL;
					break; 
				case 1:
					terrainTypes[i] = TT_WATER;
					break;
				case 2:
					terrainTypes[i] = TT_COAST;
					break;
				case 3:
					terrainTypes[i] = TT_BLOCKED;
					break;
				default:
					System.out.println ("Error: Load Map: Unknown terrain type found: " + rawData[infoPos + i]);
					return null;
				}
			}

			// Load map data
			Map map = new Map (width, height);
			for (int y = 0; y < height; y++)
			{
				for (int x = 0; x < width; x++)
				{
					int terrainIdx = ByteConvert.toInt16 (rawData, dataPos + 2 * (y * width + x));
					if (terrainIdx < 0 || terrainIdx >= numberOfTerrains)
					{
						System.out.println ("Error: Load Map: Map field has unknown terrain: " + terrainIdx);
						return null;
					}
					map.setTerrain (x, y, terrainTypes[terrainIdx]);
				}
			}
			return map;
		}
		catch (Exception e) 
		{
			System.out.println ("Can't load map: " + e);
		}
		return null;
	}

}
